<?php
// Copyright (c) 2005, RedIRIS. All Rights Reserved.
//
// You may distribute under the terms of the GNU General Public License,
// as specified in the LICENSE file that was shipped with this distribution

require_once 'crypt.php';

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS abstractPoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

abstract class abstractPoA {
	var $local_site;
	var $cfg; // Configuration parameters
	var $homeLocatorIdentifier;

	abstract function set_parameters();
	abstract function error_log($message);
	abstract function send_Request();
	abstract function check_Access();

	function abstractPoA($local_site) {
		$this->homeLocatorIdentifier = "";
	
		$this->local_site = $local_site;

		if ( ! $this->set_parameters() ) {
			trigger_error("Errors during $this configuration" , E_USER_ERROR);
			return null;
		}

		if ( ! $this->check_parameters() ) {
			trigger_error("Required parameters missing for $local_site", E_USER_ERROR);
			return null;
		}


		// Get the Simetric Key
		$fp1 = fopen($this->cfg["LKEY_File"], "r");
		$this->lkey = fread($fp1, 8192);
		fclose($fp1);
		
		// Get the Public Key
		$this->get_public_key();
	}

	function getHomeLocatorIdentifier() {
		return $this->homeLocatorIdentifier;
	}
	
	function setHomeLocatorIdentifier($hli) {
		$this->homeLocatorIdentifier=$hli;
	}

	// Taken from php.net to replace php 5 functions
	function __http_build_query($formdata, $numeric_prefix) {
		$arr = array ();
		foreach ($formdata as $key => $val)
			$arr[] = urlencode($numeric_prefix . $key) . "=" . urlencode($val);
		return implode($arr, "&");
	}

	function papi_Engine() {
		ob_start();
								
		if(isset($this->cfg["Pass_Pattern"])){
			$pattern = $this->pass_pattern();
			if($pattern){
				$this->error_log("The request matched the pass pattern '$pattern'");
				$userAssertion = $pattern;
				return array(2, $userAssertion);
			}
		}
		
		$action = array_key_exists('ACTION', $_REQUEST) ? $_REQUEST['ACTION'] : "";
		if ($action == "CHECKED") { // GpoA/AS Response
			$data = $_REQUEST['DATA'];
			list ($valid, $key, $userAssertion) = $this->test_Response($data, $this->pub_key);
			if ($valid) { // GpoA/AS has validated the request. Check the filters	
				if (!$this->test_Filters($userAssertion)) { // Data does not match  the filters
					$error_message = "Assertion $userAssertion does not match filters";
					$this->error_log($error_message );
					# If AutoPoA automatically redirect error page
					# If PoA return userAssertion
					return array(0, $userAssertion, $this->cfg['Not_Auth_Error_File']);
				}
				// not filtered, continue!
				$Lcook_cookie = $this->gen_Lcook($userAssertion);
				if (setcookie("Lcook", $Lcook_cookie, false, $this->cfg["Location"], $this->cfg["Cookie_Domain"], 0)) {
					$_COOKIE['Lcook'] = $Lcook_cookie;
					$req_s = $this->load_Request($key);
					$req = unserialize($req_s);
					if (!$req_s) {
						$error_message = "Has not been possible to reload the original request. Access not allowed";
						$this->error_log($error_message);
						# If AutoPoA automatically redirect error page
						# If PoA return userAssertion
						return array(-1, $userAssertion, $this->cfg['System_Error_File']);
					}
					$this->delete_Request($key);
					// Reload the PHP parameters of the original request
					foreach ($req as $index => $value) {
						switch ($index) {
							case "SERVER" :
								$_SERVER = $value;
								break;
							case "POST" :
								$_POST = $value;
								break;
							case "FILES" :
								$_FILES = $value;
								break;
							case "COOKIE" :
								$_COOKIE = $value;
								break;
						}
					}
					// Reload the Apache environment of the original request
					$apache_env = array (
						"QUERY_STRING",
						"REMOTE_ADDR",
						"REMOTE_PORT",
						"REQUEST_METHOD",
						"REQUEST_URI",
						"HTTP_ACCEPT",
						"HTTP_ACCEPT_ENCODING",
						"HTTP_ACCEPT_LANGUAGE",
						"HTTP_CONNECTION",
						"HTTP_HOST",
						"HTTP_REFERER",
						"HTTP_USER_AGENT",
						"DOCUMENT_NAME",
						"DOCUMENT_URI"
					);
					reset($apache_env);
					foreach ($apache_env as $value) {
						apache_setenv($value, $_SERVER[$value]);
					}
					return array(1, $userAssertion);
				} else {
					// Cookie has not been sent OK
					$error_message = "Cookie has not been correctly sent";
					$this->error_log($error_message);
					# If AutoPoA automatically redirect error page
					# If PoA return userAssertion
					return array(-1, $userAssertion, $this->cfg['Cookie_Error_File']);
				}
			} else {
				// GpoA/AS has not validated the request. Do not allow access and send ERROR page.
				$error_message = "GpoA/AS has not validated the request";
				$this->error_log($error_message);
				# If AutoPoA automatically redirect error page
				# If PoA return userAssertion
				return array(0, $userAssertion, $this->cfg['Not_Auth_Error_File']);
			}
		} else { // Direct Access from the client browser
			$Lcook = array_key_exists('Lcook', $_COOKIE) ? $_COOKIE['Lcook'] : "";
			if ($Lcook == "") { return $this->send_Request (); } 
			else {
				// Check if Lcook is correct
				list ($res, $userAssertion) = $this->test_Lcook($Lcook);
				switch ($res) {
					case 0 :
						// User not authorized
						$error_message = "User not authorized.";
						$this->error_log($error_message);
						# If AutoPoA automatically redirect error page
						# If PoA return userAssertion
						return array(0, $userAssertion, $this->cfg['Not_Auth_Error_File']);
					case 1 :
						// Cookie Lcook is correct: Update cookie
						$Lcook_cookie = $this->gen_Lcook($userAssertion);
						if (setcookie("Lcook", $Lcook_cookie, false, $this->cfg["Location"], $this->cfg["Cookie_Domain"], 0)) {
							$_COOKIE['Lcook'] = $Lcook_cookie;
							return array(1, $userAssertion);
						} else {
							// Cookie has not been updated OK
							$error_message = "Cookie has not been correctly updated.";
							$this->error_log($error_message);
							# If AutoPoA automatically redirect error page
							# If PoA return userAssertion
							return array(1, $userAssertion, $this->cfg['Cookie_Error_File']);
						}
					case 2 :
						// Lcook has expired: Request to GpoA/AS
						return $this->send_Request ();
				} // Switch
			} // else ($Lcook != "")
		} // else ($action != "CHECKED")
		ob_end_flush();
	} // Function check_Access

	final function gen_Lcook($userAssertion) {
		$Timestamp = time();
		$content = "abcd:$Timestamp:" . $this->cfg["Location"] . ":" . $this->local_site . ":$userAssertion";
		$Lcook = encrypt_AES($content, $this->lkey);
		return "$Lcook";
	}

	final function test_Lcook($Lcook) {
		// Return:
		// - 0: Lcook not valid.
		// - 1: Lcook valid.
		// - 2: Lcook expired.
		$newsource = decrypt_AES($Lcook, $this->lkey);
		$newsource = substr($newsource, 5);
		$Lcook = explode(":", $newsource, 4);
		$timestamp = $Lcook[0];
		$loc = $Lcook[1];
		$serv = $Lcook[2];
		$userAssertion = $Lcook[3]; // Code parameter of the Lcook is the userAssertion
		// Check the cookie
		if ($userAssertion == "") {
			$error_message = "test_Lcook() Function: Code parameter of Lcook is empty.";
			$this->error_log($error_message);
			return array (
				0,
				"0"
			);
		}
		if ($loc != $this->cfg["Location"]) {
			$error_message = "test_Lcook() Function: Location parameter of Lcook is not valid: $loc";
			$this->error_log($error_message);
			return array (
				0,
				"$userAssertion"
			);
		}
		if ($serv != $this->local_site) {
			$error_message = "test_Lcook() Function: Service_ID parameter of Lcook is not valid: $serv";
			$this->error_log($error_message);
			return array (
				0,
				"$userAssertion"
			);
		}
		$now = time();
		if ($timestamp + $this->cfg["Lcook_Timeout"] < $now) {
			$error_message = "test_Lcook() Function: Lcook expired for Code $userAssertion.";
			$this->error_log($error_message);
			return array (
				2,
				"$userAssertion"
			);
		}
		if ($this->test_Filters($userAssertion)) {
			return array (
				1,
				"$userAssertion"
			);
		} else {
			$error_message = "test_Lcook() Function: Code $userAssertion matches rejection cookie filter.";
			$this->error_log($error_message);
			return array (
				0,
				"$userAssertion"
			);
		}
	}

	function test_Filters($userAssertion) {
		if(isset($this->cfg["PAPI_Filter_accept"])){
			$accept = $this->cfg["PAPI_Filter_accept"];
		}
		if(isset($this->cfg["PAPI_Filter_reject"])){
			$reject = $this->cfg["PAPI_Filter_reject"];
		}
		
		$local_match = $this->match_Filters($userAssertion, $accept, $reject);

		if ($local_match) {
			return 1;
		} else {
			$error_message = "test_Filters() Function: Code $userAssertion matches rejection cookie filter.";
			$this->error_log($error_message);
			return 0;
		}
	}

	function match_Filters($userAssertion, $accept, $reject) {
		// If there are no filters,
		// a implicit accept filter is defined: $accept == ".*"
		if ($accept == "" and $reject == "") {
			return 1;
		}
		elseif ($reject == "") {
			// Accept
			if (preg_match('/' . $accept . '/', $userAssertion)) {
				return 1;
			}
		} else // $reject is not NULL
			{
			// Accept
			if ($accept) {
				if (preg_match('/' . $accept . '/', $userAssertion)) {
					return 1;
				}
			}
			// Reject
			if (preg_match('/' . $reject . '/', $userAssertion)) {
				$error_message = "match_Filters () Function: Code $userAssertion matches rejection cookie filter.";
				$this->error_log($error_message);
				return 0;
			}
		}
		$error_message = "test_Filters() Function: Code $userAssertion matches rejection cookie filter.";
		$this->error_log($error_message);
		return 0;
	}

	final function prepare_Request() {
		// should use $_SERVER in addition to $_REQUEST here?
		$req = serialize($_REQUEST);
		list ($id, $key) = $this->save_Request($req);
		if (!$id) {
			$error_message = "It has not been possible to store the original request. Access not allowed";
			$this->error_log($error_message);
			return False;
		}
		if (isset ($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on') {
			$protocol = 'https://';
		} else {
			$protocol = 'http://';
		}
		
		//Check if the request hast to be sent to a GPoA or an AS
		$gpoaurl = $this->cfg["GPoA_URL"];
		
		// Direct Access to an AS				 
		if($gpoaurl == ""){
			//Format: URL = ASURL?ATTREQ=PoA/GPoAServiceID&PAPIPOAREF=value&PAPIPOAURL=value
			$ASURL = $protocol.$_SERVER['SERVER_NAME'].":".$_SERVER['SERVER_PORT'].$_SERVER['REQUEST_URI'];
			$parameters = array(	
				'ATTREQ'=>'poaid', 	
				'PAPIPOAREF'=>$key, 
				'PAPIPOAURL'=>$ASURL,
				);
				
			//Request to AS
			$request = $this->cfg["AS_URL"].'?'.$this->__http_build_query($parameters,"");						
		} else	{
		// Direct Access to a GPoA 				 
			$PoAURL = $protocol.$_SERVER['SERVER_NAME'].":".$_SERVER['SERVER_PORT'].$_SERVER['REQUEST_URI'];
			$parameters = array (
				'ACTION' => 'CHECK',
				'DATA' => $key,
				'URL' => $PoAURL,		
				);	
			if (strcmp($this->getHomeLocatorIdentifier(),"")!=0) {
				$parameters['PAPIHLI'] = $this->getHomeLocatorIdentifier();
			}
		 
			// Request to GpoA
			$request = $this->cfg["GPoA_URL"].'?'.$this->__http_build_query($parameters,"");
		}		
		
		return $request;
	}

	final function test_Response($data, $pub_key) {
		$newsource = "";
		// Decrypt the text 
		$newsource = openssl_decrypt($data, $pub_key, $this->cfg["error_log"]);
		if(!$newsource){
			// Cannot decrypt!
			$error_message = "test_Response() Function: Cannot decrypt response, check GPoA/AS public key.";
			$this->error_log($error_message);
			return array (
				0,
				null,
				null
			);
		}
		$response = explode(":", $newsource);
		$key = array_pop($response);
		$currentTime = array_pop($response);
		$expiryTime = array_pop($response);
		$userAssertion = implode(":", $response);
		if ($response[0] == "ERROR") { // The cookie is not valid. Not authorized User.
			$error_message = "test_Response() Function: The cookie is not valid. User not authorized.";
			$this->error_log($error_message);
			return array (
				0,
				"$key",
				"$userAssertion"
			);
		} else { // Check the response (KEY,expiryTime,filters)
			$expiryTime = $response[1];
			$currentTime = $response[2];
			// Check the cookie
			if ($userAssertion == "") {
				// userAssertion is empty
				$error_message = "test_Response() Function: userAssertion is empty.";
				$this->error_log($error_message);
				return array (
					0,
					"$key",
					"$userAssertion"
				);
			} else {
				// Check if the returned key from the GpoA/AS is stored in the DataBase
				$id = dba_open($this->cfg["Request_DB"], "r", $this->cfg['DB_Type']);
				if (!$id) {
					$error_message = "test_Response() Function: dba_open $this->cfg[Request_DB] failed.";
					$this->error_log($error_message);
					return array (
						0,
						"$key",
						"$userAssertion"
					);
				}
				elseif (!(dba_exists($key, $id))) {
					// Key $key does not exist in Data Base $cfg[Req_DB]
					$error_message = "test_Response() Function: Key $key does not exist in Data Base $this->cfg[Request_DB].";
					$this->error_log($error_message);
					return array (
						0,
						"$key",
						"$userAssertion"
					);
				}
			}
			return array (
				1,
				"$key",
				"$userAssertion"
			);
		} // Check the response (KEY,expiryTime,filters)
	} // test_Response Function

	// DBM functions
	function save_Request($req) {
		// If the DB does not exist, is created.
		$id = dba_open($this->cfg["Request_DB"], "c", $this->cfg['DB_Type']);
		if (!$id) {
			$error_message = "save_Request() Function: dba_open $this->cfg[Request_DB] failed.";
			$this->error_log($error_message);
			return array (
				$id,
				""
			);
		}
		$key = mt_rand();
		dba_replace($key, $req, $id);
		dba_close($id);
		return array (
			$id,
			$key
		);
	}

	function load_Request($key) {
		$id = dba_open($this->cfg["Request_DB"], "r", $this->cfg['DB_Type']);
		if (!$id) {
			$error_message = "load_Request() Function: dba_open $this->cfg[Request_DB] failed.";
			$this->error_log($error_message);
			return "";
		}
		if (dba_exists($key, $id)) {
			$req = dba_fetch($key, $id);
		}
		dba_close($id);
		return $req;
	}

	function delete_Request($key) {
		$id = dba_open($this->cfg["Request_DB"], "c", $this->cfg['DB_Type']);
                $result = false;
		if (!$id) {
			$error_message = "delete_Request() Function: dba_open $this->cfg[Request_DB] failed.";
			$this->error_log($error_message);
		}
		if (dba_exists($key, $id)) {
			$result = dba_delete($key, $id);
		} else {
			$error_message = "delete_Request() Function: Key $key does not exist.";
			$this->error_log($error_message);
		}
		dba_close($id);
		return $result;
	}

	final function create_Hash($code, $userAssertion) {
		$userData['PAPIAuthValue'] = $code;
		
		if($code == 2){
			$userData['PAPIPassPattern'] = $userAssertion;
		}else{
			// 128 is a security limit
			$assertComponents = explode("@", $userAssertion);
			$userData['PAPIASName'] = array_pop($assertComponents);
			$userData['PAPIAssertion'] = $userAssertion;
			$pureAssertion = implode("@", $assertComponents);
			// 128 is a security limit 
			$elements = preg_split('/[\s]*,[\s]*/', $pureAssertion, 128);
			foreach ($elements as $element) {
				list ($key, $value) = preg_split('/[\s]*=[\s]*/', $element, 2);
				$userData[$key] = $value;
			}
		}

		return $userData;
	} // create_Hash()

	function check_parameters() {
		$ret = true;
		$gpoaurl = "";
		$asurl = "";
		
		$Params = Array (
			"error_log",
			"Location", "Cookie_Domain",
			"LKEY_File", "Lcook_Timeout",
//			"PAPI_Filter_accept", "PAPI_Filter_reject",
			"DB_Type", "Request_DB"
		);
		foreach ($Params as $param) {
			if ( ! isset ($this->cfg[$param])) {
				$ret = false;
				$error_message = "set_Global_Parameters() Function: Parameter $param in $this->local_site does not exist.";
				$this->error_log($error_message);
			}
		}
		
		/* Check if the GPoA_URL or the AS_URL variables have been set up but one and only one of them
		 * can be configurated at the same time
		 * Firstly we check the local sectión
		 * 
		 * 	- If only GpoA_URL or AS_URL are set up: OK
		 *	- If GPoA_URL and AS_URL are both set up: ERROR.
		 *	- If none of them is set up: check the global configuration in PAPI_Main section:
		 *			+ If GPoA_URL and AS_URL aren't set up: ERROR.
		 *			+ Only one is set up: OK
		 *			+ If GPo_AURL and AS_URL both are set up: ERROR
		*/
		if(isset($this->config_file[$this->local_site]["GPoA_URL"])){
			$gpoaurl = $this->config_file[$this->local_site]["GPoA_URL"];
		}
		
		if(isset($this->config_file[$this->local_site]["AS_URL"])){
			$asurl = $this->config_file[$this->local_site]["AS_URL"];
			
		}
		
		//Local section
		if($gpoaurl and $asurl){
			$ret = false;
			$error_message = "set_Global_Parameters() Function: Parameter GPoA_URL, or the AS_URL, in $this->local_site are set up at the same time. ";
			$this->error_log ($error_message."\n", 3,$this->cfg["error_log"]);
		}else if($gpoaurl == "" and $asurl == ""){ 
			//PAPI_Main
			$gpoaurl = $this->config_file["PAPI_Main"]["GPoA_URL"];
			$asurl = $this->config_file["PAPI_Main"]["AS_URL"];
			if(($gpoaurl == "" and $asurl == "") or ($gpoaurl and $asurl)){
				$ret = false;
				$error_message = "set_Global_Parameters() Function: Parameter GPoA_URL, or the AS_URL, in $this->local_site has to have a value, but only one them has to be set up. ";
				$this->error_log ($error_message."\n", 3,$this->cfg["error_log"]);
			}		
		}
		
		return $ret;
	}
	
	function get_public_key(){
		$asurl = "";
		$gpoaurl = "";
				
		//LOCAL_SECTION: AS_URL or GPoA_URL are set up in the local section
		if(isset($this->config_file[$this->local_site]["AS_URL"])){
			$asurl = $this->config_file[$this->local_site]["AS_URL"];
		}else{
			$gpoaurl = $this->config_file[$this->local_site]["GPoA_URL"];			
		}

		
		if($asurl)
		{	//Direct access to AS
			$fp2=fopen($this->cfg["AS_Pub_Key"],"r");
			$this->pub_key=fread($fp2,8192);
			fclose($fp2);
		}elseif ($gpoaurl){	//Access to AS through GPoA
			$fp2=fopen($this->cfg["GPoA_Pub_Key"],"r");
			$this->pub_key=fread($fp2,8192);
			fclose($fp2);
		}else{
			//MAIN_SECTION
			$asurl = $this->config_file["PAPI_Main"]["AS_URL"];
			$gpoaurl = $this->config_file["PAPI_Main"]["GPoA_URL"];
		
			if($asurl)
			{	//Direct access to AS
				$fp2=fopen($this->cfg["AS_Pub_Key"],"r");
				$this->pub_key=fread($fp2,8192);
				fclose($fp2);
			}elseif ($gpoaurl){	//Access to AS through GPoA
				$fp2=fopen($this->cfg["GPoA_Pub_Key"],"r");
				$this->pub_key=fread($fp2,8192);
				fclose($fp2);
			}	
		}
	}
	
	function pass_pattern(){
		$query = $this->getQuery();
		$patterns = explode(",", $this->cfg["Pass_Pattern"]);
		
		// Check if any pattern matches with the URI
		$uri = "$_SERVER[SERVER_NAME]$_SERVER[REQUEST_URI]";
		foreach($patterns as $pattern){
        	if (preg_match('/'.$pattern.'/',$uri)){
	          return $pattern;
    	    }
      	}
		
		// Check if any pattern matches with the GET or POST params.
		foreach($query as $val){
			foreach($patterns as $pattern){
				if(preg_match('/' . $pattern  . '/', $val)){
					return $pattern;
				}
			}
		}
		return false;
	}
	
	function getQuery()
    {
		$data = explode("&", $_SERVER['QUERY_STRING']);

        if ($_SERVER['REQUEST_METHOD'] == 'POST') {
        	$str = file_get_contents('php://input');
			
	        if ($str === false) {
              	$post = array();
            } else {
              	$post = explode("&", $str);
            }

            $data = array_merge($data, $post);
         }
		
        return $data;
    }

} //class abstract PoA

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS simplePoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
abstract class simplePoA extends abstractPoA {
	var $config_file;

	function set_parameters() {
		// Get config file name from php.ini
		$phpPoA_ini_file = get_cfg_var('phpPoA_ini_file');
		return $this->set_Global_Parameters($phpPoA_ini_file);
	}

	function test_Filters($userAssertion) {
		$accept = "";
		$reject = "";
		
		// Main filter
		if(isset($this->config_file["PAPI_Main"]["PAPI_Filter_accept"])){
			$accept = $this->config_file["PAPI_Main"]["PAPI_Filter_accept"];
		}
		if(isset($this->config_file["PAPI_Main"]["PAPI_Filter_reject"])){
			$reject = $this->config_file["PAPI_Main"]["PAPI_Filter_reject"];
		}
		$main_match = $this->match_Filters($userAssertion, $accept, $reject);
		
		// Local filter
		if(isset($this->cfg["PAPI_Filter_accept"])){
			$accept = $this->cfg["PAPI_Filter_accept"];
		}
		if(isset($this->cfg["PAPI_Filter_reject"])){
			$reject = $this->cfg["PAPI_Filter_reject"];
		}
		$local_match = $this->match_Filters($userAssertion, $accept, $reject);
		
		// Global evaluation
		if ($local_match && $main_match) {
			return 1;
		} else {
			$error_message = "test_Filters() Function: Code $userAssertion matches rejection cookie filter.";
			$this->error_log($error_message);
			return 0;
		}
	}

	function send_Request() {
		$request = $this->prepare_Request();
		if ( ! $request ) {
			# If AutoPoA automatically redirect error page
			# If PoA return userAssertion
			return array(-1, "", "Location: " . $this->cfg['System_Error_File']);
		}
		header("Location: $request");
		exit;
	}

	function set_Global_Parameters($phpPoA_ini_file) {
		// Get parameters from config file
		$this->config_file = parse_ini_file($phpPoA_ini_file, true);
		// PAPI_Main Section
		$MAIN_Params = Array (
			"Lcook_Timeout",
			"Request_DB",
			"DB_Type",
			"error_log",
			"Not_Auth_Error_File",
			"Cookie_Error_File",
			"System_Error_File",
			"Config_Error_File",
			"PAPI_Filter_accept",
			"PAPI_Filter_reject",
			"Pass_Pattern",
			"Cookie_Domain",
			"LKEY_File",
			"GPoA_Pub_Key",
			"GPoA_URL",
			"AS_Pub_Key",
			"AS_URL"
		);
		if (!isset ($this->config_file["PAPI_Main"]["Cookie_Domain"])) {
			$this->config_file["PAPI_Main"]["Cookie_Domain"] = $_SERVER["SERVER_NAME"];
		}
		foreach ($MAIN_Params as $param) {
			if(isset($this->config_file["PAPI_Main"][$param])){
				$this->cfg[$param] = $this->config_file["PAPI_Main"][$param];
			}
		}
		// The parameters in local section of phpPoA_ini_file have priority over parameters in the PAPI_Main section with the same name. Therefore the local parameters set their own value over global parameters. As the filter parameters must not have priority because we have to check both of them, to access PAPI_Filter_accept and PAPI_Filter_reject from PAPI_Main Section we have to use:
		// $config_file["PAPI_Main"]["PAPI_Filter_accept"]
		// $config_file["PAPI_Main"]["PAPI_Filter_reject"]
		// instead of $cfg.
		// PAPI_Local Section
		$LOCAL_Params = Array (
			"Location",
			"Lcook_Timeout",
			"Request_DB",
			"DB_Type",
			"error_log",
			"Not_Auth_Error_File",
			"Cookie_Error_File",
			"System_Error_File",
			"Config_Error_File",
			"Cookie_Domain",
			"PAPI_Filter_accept",
			"PAPI_Filter_reject",
			"Pass_Pattern",
			"LKEY_File",
			"GPoA_Pub_Key",
			"GPoA_URL",
			"AS_Pub_Key",
			"AS_URL"
		);
		if (!isset ($this->config_file[$this->local_site]["Cookie_Domain"])) {
			$this->config_file[$this->local_site]["Cookie_Domain"] = $_SERVER["SERVER_NAME"];
		}
		foreach ($LOCAL_Params as $param) {
			if (isset ($this->config_file[$this->local_site][$param])) {
				$this->cfg[$param] = $this->config_file[$this->local_site][$param];
			}
		}
		return True;
	} // set_Global_Parameters()

	function error_log($message) {
		$error_message = date("d-M-Y H:i:s") . " " . $this->local_site . " : " . $message;
		error_log($error_message . "\n", 3, $this->cfg["error_log"]);
	}
} //class simplePoA

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS PoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class PoA extends simplePoA {

	function check_Access() {
		list($code, $userAssertion) = $this->papi_Engine();
		$userData = $this->create_Hash($code, $userAssertion);
		return $userData;
	}
} // Class PoA


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS autoPoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class autoPoA extends simplePoA {
	function check_Access() {
		list($code, $userAssertion, $location) = $this->papi_Engine();
		if ($location && $code < 1) // error found, redirect!
			header("Location: " . $location);
		else // this shouldn't be an error, but can be. Anyway, do not redirect!
			return $this->create_Hash($code, $userAssertion);
	}
}
?>
